#!/usr/bin/env python

from __future__ import print_function
import argparse
import datetime
import logging
from multiprocessing import Manager, Process, Lock
import os
import os.path
import requests
from requests.packages.urllib3 import disable_warnings
from requests.packages.urllib3.exceptions import InsecureRequestWarning
import sys
from time import sleep
import yaml

"""
SPDX-License-Identifier: BSD-2-Clause
Copyright (c) 2020 Panzura, Inc. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
       notice, this list of conditions and the following disclaimer.
    2. Redistributions in binary form must reproduce the above copyright
       notice, this list of conditions and the following disclaimer in the
       documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

__version_info__ = ('2', '0', '0')
__version__ = '.'.join(__version_info__)

starttime = datetime.datetime.now()

# yaml and requests are third party packages that need to be
# installed for this script to run

# Things we need for coordination between the processes
manager = Manager()
glist = manager.list()
gpb = manager.list()
gpb.append(0)
gpb.append(0)
gkill = manager.list()
lock = Lock()


def load_config(configfile):
    if not configfile:
        configfile = 'apisetup.yaml'

    if os.path.isfile(configfile):
        try:
            with open(configfile, 'r', encoding='utf8') as f:
                fh = f.read()
        except TypeError:
            with open(configfile, 'r') as f:
                fh = f.read()
        except IOError as e:
            print("I/O error({0}): {1}".format(e.errno, e.strerror))
            sys.exit(1)
        try:
            cfg = yaml.load(fh, Loader=yaml.FullLoader)
        except AttributeError:
            cfg = yaml.load(fh)
        except yaml.YAMLError as e:
            sys.exit("Aborting due to config file YAML error.")
    else:
        sys.exit("Config file not found.")
    return cfg


# disable the requests module SSL cert verification
disable_warnings(InsecureRequestWarning)


def check_and_mkdir(path):
    """Helper function to create a dir if it doesn't exist"""
    if os.path.exists(path) and not os.path.isdir(path):
        sys.exit("%s exists but is not a directory" % path)
    elif not os.path.exists(path):
        try:
            os.mkdir(path)
        except OSError:
            sys.exit("mkdir %s failed." % path)
    return True


def mk_logdir(path, supress=False):
    """Create a timestamped log dir that will presumably be
    unique every run.  There will be a logfile per cluster
    member in this directory."""
    if path:
        logbase = path
    else:
        logbase = 'log'
    check_and_mkdir(logbase)

    logdir_ts = str(int((starttime - datetime.datetime(1970, 1, 1)).total_seconds()))
    logdir = os.path.join(logbase, logdir_ts)
    check_and_mkdir(logdir)

    if not supress:
        print("logdir: %s" % logdir)
    return logdir


def updt(total, progress):
    """
    Displays or updates a console progress bar.
    """
    if progress > total:
        return
    barLength, status = 20, ""
    progress = float(progress) / float(total)
    if progress >= 1.:
        progress, status = 1, "\r\n"
    block = int(round(barLength * progress))
    text = "\r[{}] {:.0f}% {}".format(
            "#" * block + "-" * (barLength - block),
            round(progress * 100, 0), status)
    sys.stdout.write(text)
    sys.stdout.flush()


def call_updt():
    updt(gpb[0], gpb[1] + 1)
    gpb[1] += 1


class config():
    def __init__(self, host, role, logdir, cfg):
        self.host = host
        self.role = role
        self.logdir = logdir
        self.cfg = cfg
        # layering violation.  We can't use getkey() yet
        self.base = 'https://%s:443/' % self.cfg['cloudfsmember'][self.host]['ip']
        self.headers = {'accept': 'application/json',
                        'content-type': 'application/json'}
        # Configure the per host logger
        self.logger = logging.getLogger(self.host)
        self.logfile = os.path.join(self.logdir, self.host)
        self.logger.setLevel(logging.DEBUG)
        self.c_handler = logging.StreamHandler()
        self.f_handler = logging.FileHandler(self.logfile)
        self.c_handler.setLevel(logging.INFO)
        self.f_handler.setLevel(logging.DEBUG)
        self.c_format = logging.Formatter('%(name)s: %(message)s')
        self.f_format = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        self.c_handler.setFormatter(self.c_format)
        self.f_handler.setFormatter(self.f_format)
        self.logger.addHandler(self.c_handler)
        self.logger.addHandler(self.f_handler)

    def getkey(self, key):
        """look for host specific config, then fall back
         to a global config"""
        if self.host and key in self.cfg['cloudfsmember'][self.host]:
            return self.cfg['cloudfsmember'][self.host][key]
        else:
            return self.cfg['global'][key]

    def apicall(self, op, url, data, message, datareturn):
        """Wrapper around requests to make an api call."""
        self.logger.debug("Enter: %s" % message)
        # op can be put, patch, post, or get, and generate the
        # correct requests.op function accordingly.
        # Be super defensive here as op could be set to anything.
        try:
            r_func = getattr(requests, op)
        except AttributeError as e:
            gkill.append(self.host)
            sys.exit(e)
        try:
            assert callable(r_func)
        except AssertionError:
            gkill.append(self.host)
            sys.exit("Falling on my sword now")

        # Try to handle non JSON being passed in for data
        if isinstance(data, str):
            r = r_func(url, headers=self.headers, data=data, verify=False)
        else:
            r = r_func(url, headers=self.headers, json=data, verify=False)

        ret = None
        if r.ok:
            self.logger.debug("Exit: %s: Success" % message)
            if datareturn:
                n = r.json()
                if datareturn is True:
                    ret = n
                else:
                    ret = n[datareturn]
            return (True, ret)
        else:
            self.logger.debug("Exit: %s: Failed" % message)
            self.logger.debug("%s %s %s" % (r.status_code, r.reason, r.json()))
            return (False, r.json())

    def isalive(self):
        """Check that the host is alive by pinging it."""
        count = 0
        while True:
            if count % 5:
                self.logger.debug("isalive waiting")
            ret = os.system('ping -c 1 %s > /dev/null 2>&1' % self.getkey('ip'))
            if not ret:
                self.logger.debug("is alive")
                break
            count += 1
        sleep(5)
        call_updt()

    def getapikey(self):
        """Get an API key"""
        endpoint = 'pzapi/auth/login'
        url = self.base + endpoint

        data = {'username': self.getkey('username'),
                'password': self.getkey('password')}
        try:
            ret, apikey = self.apicall('post', url, data, "Retrieving API Key", 'apiKey')
        except Exception as e:
            print(e)
            gkill.append(self.host)
            sys.exit("Exception getting apikey")
        if not ret or not apikey:
            gkill.append(self.host)
            sys.exit("Error getting apikey")

        call_updt()
        return apikey

    def seteula(self):
        endpoint = 'pzapi/tool/eula'
        url = self.base + endpoint
        ret, val = self.apicall('put', url, self.getkey('eula'), "Accepting EULA", False)
        call_updt()
        if not ret:
            # non fatal error
            self.logger.debug(val)

    def preconfig(self):
        data = {}
        for item in ['domainname', 'nameServers',
                     'ntpMode', 'ntpServer',
                     'systemLocation']:
            data[item] = self.getkey(item)
        endpoint = 'pzapi/config'
        url = self.base + endpoint
        ret, val = self.apicall('put', url, data, "Setting Basic info", False)
        call_updt()
        if not ret:
            self.logger.info("Aborting cluster setup")
            gkill.append(self.host)

    def setrole(self, role):
        data = {'hostname': self.host,
                'domainname': self.getkey('domainname'),
                'private_secure_site_mode': self.getkey('private_secure_site_mode')}
        roleitems = self.getkey('role')
        for item in list(roleitems.keys()):
            data[item] = roleitems[item]
        endpoint = 'pzapi/config/configureRole'
        url = self.base + endpoint

        lock.acquire()
        i = 0
        while True:
            ret, val = self.apicall('put', url, data, "Configure %s Role" % role, False)
            if not ret:
                if i < 5:
                    i += 1
                    sleep(5)
                    continue
                else:
                    self.logger.info("Aborting cluster setup")
                    gkill.append(self.host)
            lock.release()
            break
        call_updt()

    def keymaster(self):
        data = {'keyMasterEnable': True}
        endpoint = 'pzapi/config'
        url = self.base + endpoint
        ret, val = self.apicall('put', url, data, "Enable key master", False)
        call_updt()
        if not ret:
            self.logger.info("Aborting cluster setup")
            gkill.append(self.host)

    def genkey(self):
        endpoint = 'pzapi/pairingkey'
        url = self.base + endpoint
        data = {}
        ret, val = self.apicall('post', url, data, "Generate Pairing Key", 'value')
        call_updt()
        if not ret:
            self.logger.info("Aborting cluster setup")
            gkill.append(self.host)
        else:
            glist.append(val)

        # Export paring key on master.
        ret, val = self.apicall('get', url, data, "Export Pairing Key", 'result')
        if not ret:
            self.logger.info("Aborting cluster setup")
            gkill.append(self.host)
        else:
            glist.append(val)

    def putkey(self):
        data = {"result": glist[1]}
        endpoint = 'pzapi/pairingkey'
        url = self.base + endpoint
        ret, val = self.apicall('post', url, data, "Import Pairing Key", 'value')
        call_updt()
        if not ret:
            self.logger.info("Aborting cluster setup")
            gkill.append(self.host)
        elif val != glist[0]:
            self.logger.info("Aborting cluster setup")
            gkill.append(self.host)

    def createdisks(self):
        endpoint = 'pzapi/disk/autoCreate'
        url = self.base + endpoint
        ret, val = self.apicall('put', url, self.getkey('disk_data'), "Creating VM disks", False)
        call_updt()
        if not ret:
            self.logger.info("Aborting cluster setup")
            gkill.append(self.host)

    def addlicense(self):
        # Activate license token and csp license.
        endpoint = 'pzapi/license/token'
        url = self.base + endpoint
        lock.acquire()
        while True:
            ret, val = self.apicall('put', url, self.getkey('lic_data'),
                                    "Uploading and Configuring Licenses", False)
            if not ret:
                sleep(5)
                continue
            break
        lock.release()
        call_updt()

    def rundbg(self):
        # spin on pzapi/tool/rundbg until it returns Bootup Complete
        endpoint = 'pzapi/tool/rundbg'
        url = self.base + endpoint
        self.logger.debug("Waiting for system to be ready")
        count = 0
        while True:
            self.logger.debug("waiting")
            if count < 21:
                call_updt()
                sleep(.05)
            try:
                r = requests.get(url, headers=self.headers, verify=False)
            except requests.exceptions.RequestException as e:
                self.logger.debug(e)
                sleep(30)
                continue
            if r.ok:
                if "Bootup Complete" in r.text:
                    # ok to continue
                    self.logger.debug("Wait Complete")
                    break
                else:
                    count += 1
                    sleep(5)
            else:
                # Something went really wrong
                self.logger.debug("System returned an unexpected result while "
                                  "waiting to come ready.")
                self.logger.debug("%s %s" % (r.status_code, r.reason))
                self.logger.info("Aborting cluster setup")
                gkill.append(self.host)
        while True:
            if count < 21:
                call_updt()
                count += 1
            else:
                break

    def setupdone(self):
        # Set cc setup Done
        data = {}
        endpoint = 'pzapi/tool/setupDone'
        url = self.base + endpoint
        ret, val = self.apicall('put', url, data, "Completing Setup", False)
        if not ret:
            # non fatal error
            pass
        call_updt()

    def startservices(self):
        data = {"clientIfMode": self.getkey('clientIfMode'),
                "cifsMode": self.getkey('cifsMode'),
                "nfsMode": self.getkey('nfsMode')}
        while True:
            endpoint = 'pzapi/config'
            url = self.base + endpoint
            ret, val = self.apicall('put', url, data, "Enable SMB and NFS", False)
            if not ret:
                if "CIFS service is in starting, please wait." in val['errors']:
                    sleep(5)
                else:
                    # not the end of the world if CIFS doesn't start
                    pass
            else:
                break
        call_updt()

    def adserver(self):
        while True:
            endpoint = 'pzapi/config'
            url = self.base + endpoint
            ret, val = self.apicall('put', url, self.getkey('adServer'),
                                    "Configure CC AD server", False)
            if not ret:
                if val['status'] == 400:
                    sleep(3)
            else:
                break
        call_updt()

    def adjoin(self):
        data = {'username': self.getkey('adusername'),
                'password': self.getkey('adpassword')}
        endpoint = 'pzapi/tool/adjoin'
        url = self.base + endpoint
        ret, val = self.apicall('post', url, data, "Joining AD", False)
        if not ret:
            self.logger.info("Aborting cluster setup")
            gkill.append(self.host)

        ret, val = self.apicall('get', url, data, "Confirming AD Join", 'result')
        if not ret:
            self.logger.info("Aborting cluster setup")
            gkill.append(self.host)
        call_updt()

    def stageconfig(self, stage, obj):
        endpoint = 'pzapi/config'
        url = self.base + endpoint
        ret, val = self.apicall('put', url, obj[stage],
                                "Uploading %s Config" % stage, False)
        if not ret:
            self.logger.info("Aborting cluster setup")
            gkill.append(self.host)
        call_updt()

    def setpassword(self):
        data = {'password': self.getkey('newPassword'),
                'originalPassword': self.getkey('password'),
                'adminOrUser': self.getkey('username')}
        endpoint = 'pzapi/auth/changePassword'
        url = self.base + endpoint
        ret, val = self.apicall('post', url, data, "Change Password", False)
        call_updt()

    def do_config(self):
        call_updt()
        self.isalive()
        self.headers['X-API-Key'] = self.getapikey()
        self.seteula()
        self.preconfig()

        if self.role == 'master':
            self.setrole(self.role)
            self.keymaster()
            self.genkey()

        else:
            # role is either subordinate or ha_*
            self.putkey()
            self.setrole(self.role)

        self.createdisks()
        self.addlicense()
        self.rundbg()
        self.setupdone()

        if self.role == 'master':
            self.startservices()

        self.adserver()
        self.adjoin()

        # stageconfig is broken up so the API doesn't
        # time out
        for stage in self.cfg['stageconfig']:
            self.stageconfig(stage, self.cfg['stageconfig'])

        if self.role == 'master':
            self.setpassword()

        self.logger.debug("Host setup complete.")


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="cluster setup script")
    parser.add_argument('-l', '--logdir', default=None,
                        help="logging directory (defaults to ./log)")
    parser.add_argument('-c', '--config', default=None,
                        help="config file, defaults to ./apisetup.yaml")
    parser.add_argument('-V', '--version', action='version',
                        version='%(prog)s {version}'.format(version=__version__))
    args = parser.parse_args()
    logdir = mk_logdir(args.logdir, False)
    cfg = load_config(args.config)

    host_list = {'master': [],
                 'subordinate': [],
                 'ha_local': [],
                 'ha_global': []}

    roles = list(host_list.keys())
    for host in cfg['cloudfsmember']:
        for role in roles:
            if cfg['cloudfsmember'][host]['role']['role'] == role:
                host_list[role].append(host)

    # Do some sanity checking
    assert len(host_list['master']) == 1
    assert len(host_list['ha_local']) <= (len(host_list['master']) + len(host_list['subordinate']))

    # magic numbers needed to make the progress bar
    # work out ok
    progbar_total = {'master': 44,
                     'subordinate': 40,
                     'ha': 40}

    # iterate explicitly because order matters
    # TODO: OrderedDict
    for system_type in ['master', 'subordinate', 'ha']:
        gpb[0] = 0
        gpb[1] = 0
        s = {}
        print()
        print("Configuring %s(s): " % system_type, end=' ')
        # conflate ha_local and ha_global here because systems of either
        # type can be configured in parallel
        if system_type == 'ha':
            host_list[system_type] = host_list['ha_local'] + host_list['ha_global']
        for host in host_list[system_type]:
            print(host, end=' ')
        print()
        sys.stdout.flush()
        for host in host_list[system_type]:
            gpb[0] += progbar_total[system_type]
            if system_type == 'ha':
                if host in host_list['ha_local']:
                    cc = config(host, 'ha_local', logdir, cfg)
                else:
                    cc = config(host, 'ha_global', logdir, cfg)
            else:
                cc = config(host, system_type, logdir, cfg)
            s[host] = Process(target=cc.do_config, args=())
            s[host].start()
        for ds in list(s.keys()):
            s[ds].join()
        if len(gkill) > 0:
            sys.exit("%s fatal error aborted cluster setup" % gkill)
        while True:
            if gpb[1] < gpb[0]:
                call_updt()
                sleep(.05)
            else:
                break

        print()
        print("%s config done" % system_type)

    endtime = datetime.datetime.now()
    print("Cluster setup complete in: %s" % (endtime - starttime))
